"""
When setting up a Docker in Snowflake Container Services
we discovered that Snowflake doesn't like using an RSA key for authentication.
Instead, Snowflake Containers make stuff available via the os environment variables
"""
import os
import functools
from snowflake.connector import connect
from sqlalchemy import create_engine


def get_container_login_token():
    with open('/snowflake/session/token', 'r') as f:
        return f.read()


def get_container_connection():
    return connect(
        host=os.getenv('SNOWFLAKE_HOST'),
        account=os.getenv('SNOWFLAKE_ACCOUNT'),
        token=get_container_login_token(),
        authenticator='oauth'
    )


def snowflake_container_sql_engine():
    """ Connection to Snowflake via the SQL Alchemy Engine Method. """
    return create_engine("snowflake://not@used/db", creator=get_container_connection)


def container_connect_decorator(func):
    """
    Snowflake Container Service Connection Wrapper

    Designed to wrap around snowflake_connect when we may use a Snowflake Container Service Connection.
    Point is to check the os.environ to see if we are in a container service.
    If we are then run a specialist container functions rather than the original function.
    """

    # the extra decorator means we can still use help etc.
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # we know if we are in a Snowflake container if 'SNOWFLAKE_HOST' in the os.environ
        if 'SNOWFLAKE_HOST' in os.environ.keys():
            return get_container_connection().raw_connection()
        else:
            return func(*args, **kwargs)

    return wrapper


def container_engine_decorator(func):
    """
    Snowflake Container Service Engine Wrapper

    Designed to wrap around snowflake_sql_engine when we may use a Snowflake Container Service Connection.
    Point is to check the os.environ to see if we are in a container service.
    If we are then run a specialist container functions rather than the original function.
    """

    # the extra decorator means we can still use help etc.
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # we know if we are in a Snowflake container if 'SNOWFLAKE_HOST' in the os.environ
        if 'SNOWFLAKE_HOST' in os.environ.keys():
            return snowflake_container_sql_engine().raw_connection()
        else:
            return func(*args, **kwargs)

    return wrapper



import pandas as pd
from snowflake.connector.pandas_tools import write_pandas


def snowflake_create_or_replace(df, conn, table_name: str = "TEST", schema: str = "SNOWBERG"):
    """ Simple Function to Create New or Replace Existing Table in Snowflake

    Careful! Replacing an existing table will delete the old version and lose all original data

    INPUTS:
        df: pd.DataFrame
        conn: snowflake connect object; best to use our snowflake_connect method to create it
        table_name: str: name required in snowflake
        schema [optional]: can provide the schema in snowflake if it wasn't part of the connection object

    """

    assert isinstance(table_name, str), f"table_name must be type string: type {type(table_name)} provided"
    assert isinstance(df, pd.DataFrame), f"""
        Can only create or replace dataframes, for {table_name}: type{type(df)}-passed """

    with conn.cursor() as cur:

        # use the show tables in SQL to find table of that name; use that to tell if the table exists
        cur.execute(f"SHOW TABLES LIKE '{table_name}'")
        rows = cur.fetchall()
        if len(rows) > 0:
            cur.execute(f""" DROP TABLE {table_name}""")

        write_pandas(conn, df, table_name, auto_create_table=True)
        conn.commit()
        print(f"{table_name} pushed to Snowflake")


if __name__ == "__main__":
    print("Foo, Bar, Baz")



""" SQL Helper Functions / Recipes """
import pandas as pd
from snowflake.connector.pandas_tools import write_pandas


def posh_create_or_append(conn, df: pd.DataFrame, table_name: str, cols: tuple | list | set = ()):
    """ Creates a new table or appends, removing duplicated with grouper columns """

    assert isinstance(df, pd.DataFrame), f"df needs to by a pandas dataframe: type{type(df)}-passed"
    assert isinstance(table_name, str), f"""
        table_name must be a string of an existing table or the one to be created: {type(table_name)}-passed """

    with conn.cursor() as cur:

        # use the show tables in SQL to find table of that name; use that to tell if the table exists
        cur.execute(f"SHOW TABLES LIKE '{table_name}'")
        rows = cur.fetchall()
        if len(rows) > 0:
            sql_append_with_drop_duplicates(conn, df, table_name, cols)
        else:
            write_pandas(conn, df, table_name, auto_create_table=True)


def posh_create_or_replace(conn, df: pd.DataFrame, table_name: str):
    """ Creates a new table or replaces """

    assert isinstance(df, pd.DataFrame), f"df needs to by a pandas dataframe: type{type(df)}-passed"

    with conn.cursor() as cur:

        # use the show tables in SQL to find table of that name; use that to tell if the table exists
        cur.execute(f"SHOW TABLES LIKE '{table_name}'")
        rows = cur.fetchall()
        if len(rows) > 0:
            cur.execute(f""" DROP TABLE {table_name}""")

        write_pandas(conn, df, table_name, auto_create_table=True)
        conn.commit()


def check_compatibility(conn, df: pd.DataFrame, table_name: str, map_fudge: dict = {}):
    """ Snowflake Helper Function to Check Dataframe Compatability Before Upload

    In testing had some problems with datatypes when uploading data to Snowflake from a Pandas Dataframe.
    Function runs a series of assert statements to check for compatability.

    Asserts:
    - number of columns in df and snowflake table are the same
    - column names in df match table names in snowflake table
    - data types are mapped; this may not be perfect.

    :arg:
        conn: snowflake connection object
        df: pd.DataFrame: dataframe to be appended
        table_name: str: the table name in Snowflake / the SQL databass
        map_fudge: dict: way of adding data types that may not have been included when the function was written.

    :returns: bool == True if no assertion errors"""

    # get a pd.Series of df data types
    df_types = pd.Series(df.dtypes)

    # Get the column names and data types of the Snowflake table
    with conn.cursor() as cur:
        cur.execute(f"""DESCRIBE {table_name}""")
        result = cur.fetchall()
        cols = [col[0] for col in cur.description]
        original = pd.DataFrame(result, columns=cols).set_index('name')

    # df we want the width, which is no of columns... hence [1]
    # originals df has table cols listed per row... hence [0]
    assert df.shape[1] == original.shape[0], \
        f"Column count in new df: {df.shape[1]} does not match snowflake table: {original.shape[0]}"

    # attempt to map types between Snowflake & Pandas
    # Ref: https://docs.snowflake.com/en/developer-guide/python-connector/python-connector-pandas
    map_snowflake_types = {
        'DATE': ['object'],
        'TIME': ['object'],
        'TIMESTAMP_NTZ': ['datetime64[ns]'],
        'TIMESTAMP_LTZ': ['datetime64[ns]'],
        'TIMESTAMP_TZ': ['datetime64[ns]'],
        'FLOAT': ['float64'],
        'VARCHAR(16777216)': ['object', 'str'],
    }

    # quite possible we are missing dtypes here
    # to avoid having to update hxm-refuel each time we find one, option to append additional types
    map_snowflake_types = map_snowflake_types | map_fudge

    # iterate over each column from the Snowflake Table
    for i in original.index:
        snowflake_type = original.loc[i, 'type']  # snowflake dtype

        assert i in df.columns, \
            f""" Snowflake Table Col {i} missing from dataframe """
        assert snowflake_type in map_snowflake_types.keys(), \
            f""" Snowflake type {snowflake_type} missing from snowflake-to-pandas map. 
            Doesn't necessarily mean something wrong, but update the map to continue. """
        assert df_types[i] in map_snowflake_types[snowflake_type], \
            f""" For col: {i} df has type {df_types[i]}. Snowflake table type {snowflake_type}; 
            this Snowflake type should map to pd types {map_snowflake_types[snowflake_type]}"""

    return True


def sql_append_with_drop_duplicates(
        conn,
        df: pd.DataFrame,
        table_name: str,
        cols: list | set | tuple,
        delete_stage: bool = True,
        run_check_compatability: bool = True):
    """ Append to Snowflake Table but Dropping Duplicates

    Hat tip to Jaco, because this was basically his idea so get the data in via a staging area.
    What we want is an equivalent to df.groupby(cols).last() but in Snowflake, which has weird indexing.
    Approach is to upload to staging table, find duplicates by column-group and remove from the original table,
    then append the new dataframe.

    """

    assert type(cols) in (list, set, tuple), f"cols must be type list|set|tuple but type {cols}-passed"

    if run_check_compatability:
        assert check_compatibility(conn, df, table_name)

    # setup context handler for cursor object
    with conn.cursor() as cur:

        # create staging table & append dataframe
        # in testing trying to create or replace within write_pandas was very hit & miss
        cur.execute(f"""CREATE OR REPLACE TABLE {table_name}_STAGE LIKE {table_name}""")
        write_pandas(conn, df, f"{table_name}_STAGE")  # alternative method that doesn't use engine

        # groupby and delete from original query
        query_delete = f"""
            DELETE FROM {table_name}
            WHERE ({", ".join(f'"{col}"' for col in cols)})
            IN (SELECT {", ".join(f'"{col}"' for col in cols)} FROM {table_name}_STAGE)
            """

        # run delete query & insert staging table into original
        cur.execute(query_delete)
        cur.execute(f"""INSERT INTO {table_name} SELECT * FROM {table_name}_STAGE""")

        # optional to drop the staging table from the database
        if delete_stage:
            cur.execute(f""" DROP TABLE {table_name}_STAGE""")

        conn.commit()  # commit changes
        return True



""" Data Validation Functions """

# %%
import types
from warnings import warn
import numpy as np
import pandas as pd
import inspect
from functools import wraps


# %% Static Methods

def _validate_no_nans(inputs: dict, name: str):
    """ Raise ValueError if any Input is NaN """

    # iterate over inputs dictionary & test if value is a NaN
    for k, v in inputs.items():

        # error message
        msg = f"{name} error: input {k} is {v}; nans not allowed"

        if pd.api.types.is_number(v):
            if np.isnan(v).any():
                raise ValueError(msg)


def _test_class(var, hint, value, name):
    """ Test value vs. hint which hint was <class 'int'> etc... """
    if not type(value) == hint:
        msg = f"{name} error: var {var} of type {type(value)} passed; " \
              f"doesn't match function type hint == {hint}"
        raise ValueError(msg)


def _test_union_hint(var, hint, value, name):
    """ Test value vs. hint when the hint is a types.UnionType """

    # one can't just convert UnionType back to a tuple of types - need a hack!
    # we convert to string, remove whitespaces then split on the union (|)
    deunion = f"{hint}".replace(" ", "").split("|")

    # now iterate over union
    for i, v in enumerate(deunion):
        match v:
            case 'None':
                deunion[i] = types.NoneType
            case 'int':
                deunion[i] = int
            case 'float':
                deunion[i] = float
            case 'str':
                deunion[i] = str
            case 'dict':
                deunion[i] = dict
            case 'list':
                deunion[i] = list
            case 'tuple':
                deunion[i] = tuple
            case 'numpy.ndarray':
                deunion[i] = np.ndarray
            case 'pandas.core.frame.DataFrame':
                deunion[i] = pd.DataFrame
            case 'pandas.core.series.Series':
                deunion[i] = pd.Series
            case 'pandas._libs.tslibs.timestamps.Timestamp':
                deunion[i] = pd.Timestamp
            case _:
                # if un-coded type, throw a warning
                deunion[i] = types.NoneType
                msg = f"data validation warning: un-coded type {v} in union "
                warn(msg)

    # we zip tuples of (value, type) i.e. (7.5, float)
    # iterate over with isinstance to get a list of booleans
    # if any bool is True means we had a type match
    z = any([isinstance(*z) for z in zip([value] * len(deunion), deunion)])
    if not z:
        msg = f"{name} error: var {var} of type {type(value)} passed; " \
              f"doesn't match function type hint == {hint}"
        raise ValueError(msg)


# %% Data Validation Class

class TypeHintValidation:
    """ TopGun Type Hint Based Data Validation Class

    Idea is to use as decorator to functions where we want to validate inputs.
    By default, this will validate that inputs don't breach function type hints.
    There is also an option to apply additional validation 'methods' by adding
    these as list.

    We can also apply additional methods:
        - nan: will raise error if a number type has had a NaN value is passed
        - others haven't been written

    Important:
        - Decorator ONLY receives explicitly input variables!
        - means a func default value has a type error it will not be checked.

    Example Syntax:
    @TypeHintValidation(methods=['nan'])
    def f(x: int, y: pd.Series | pd.DataFrame):

    References:
        https://medium.com/@ankurpan96/a-handy-validator-using-decorators-in-python-a8722da02fba
        https://levelup.gitconnected.com/validations-in-python-using-metaprogramming-and-decorators-advanced-python-ee4d4278a6b3
        https://stackoverflow.com/questions/10176226/how-do-i-pass-extra-arguments-to-a-python-decorator
    """

    def __init__(self, methods=None):

        if methods is None:
            self.methods = []
        elif isinstance(methods, str):
            self.methods = [methods]
        elif isinstance(methods, list):
            self.methods = methods
        else:
            self.methods = []
            msg = f"data validation warning: methods = {methods} provided, " \
                  f"which is unknown; additional methods set to None."
            warn(msg)

    def __call__(self, func):

        # add func to self, so we can access it in the inner function
        self.func = func
        self.hints = func.__annotations__   # dict calling func type hints

        @wraps(func)
        def data_validation_function(*args, **kwargs):
            """ Inner Function in TypeHintValidation Decorator """

            # create dictionary or inputs where var name is the key
            inputs = dict(zip(self.func.__code__.co_varnames, args)) | kwargs

            # iterate over inputs
            # remember only user inputs passed; func defaults not given to decorator
            for k, v in inputs.items():

                # frequently None is given as an input not match by Type Hine
                # generally where a func grabs something from self in a class
                # we provide option to skip tests if None is provided
                if 'none' in self.methods and isinstance(v, types.NoneType):
                    continue

                # check if k-th var name has type hint
                if k in self.hints:

                    hint = self.hints[k]    # type hint object for k-th input

                    # hints can come in multiple formats
                    # most commonly hint will be class of type i.e <class 'int'>
                    # these are trivial to unpack using inspect.isclass()
                    if inspect.isclass(hint):
                        _test_class(k, hint, v, self.func.__name__)
                    elif isinstance(hint, types.NoneType):
                        # for None, we need to pass types.NoneType as hint
                        _test_class(k, types.NoneType, v, self.func.__name__)
                    elif isinstance(hint, types.UnionType):
                        _test_union_hint(k, hint, v, self.func.__name__)
                    else:
                        warn(f"data validation: unknown method in {hint}")

            # additional methods:
            if 'nan' in self.methods:
                _validate_no_nans(inputs, self.func.__name__)

            return self.func(*args, **kwargs)

        return data_validation_function
